package ru.dbotthepony.kbox2d.api

import ru.dbotthepony.kvector.vector.ndouble.Vector2d

/// The body type.
/// static: zero mass, zero velocity, may be manually moved
/// kinematic: zero mass, non-zero velocity set by user, moved by solver
/// dynamic: positive mass, non-zero velocity determined by forces, moved by solver
enum class BodyType {
	STATIC,
	KINEMATIC,
	DYNAMIC
}

/// A body definition holds all the data needed to construct a rigid body.
/// You can safely re-use body definitions. Shapes are added to a body after construction.
data class BodyDef(
	/**
	 * The world position of the body. Avoid creating bodies at the origin
	 * since this can lead to many overlapping shapes.
 	 */
	var position: Vector2d = Vector2d.ZERO,

	/**
	 * The world angle of the body in radians.
	 */
	var angle: Double = 0.0,

	/**
	 * The linear velocity of the body's origin in world co-ordinates.
	 */
	var linearVelocity: Vector2d = Vector2d.ZERO,

	/**
	 * The angular velocity of the body.
	 */
	var angularVelocity: Double = 0.0,

	/**
	 * Linear damping is use to reduce the linear velocity. The damping parameter
	 * can be larger than 1.0f but the damping effect becomes sensitive to the
	 * time step when the damping parameter is large.
	 * Units are 1/time
	 */
	var linearDamping: Double = 0.0,

	/**
	 * Angular damping is use to reduce the angular velocity. The damping parameter
	 * can be larger than 1.0f but the damping effect becomes sensitive to the
	 * time step when the damping parameter is large.
	 * Units are 1/time
 	 */
	var angularDamping: Double = 0.0,

	/**
	 * Set this flag to false if this body should never fall asleep. Note that
	 * this increases CPU usage.
 	 */
	var allowSleep: Boolean = true,

	/**
	 * Is this body initially awake or sleeping?
	 */
	var awake: Boolean = true,

	/**
	 * Should this body be prevented from rotating? Useful for characters.
	 */
	var fixedRotation: Boolean = false,

	/**
	 * Is this a fast moving body that should be prevented from tunneling through
	 * other moving bodies? Note that all bodies are prevented from tunneling through
	 * kinematic and static bodies. This setting is only considered on dynamic bodies.
	 * @warning You should use this flag sparingly since it increases processing time.
	 */
	var bullet: Boolean = false,

	/**
	 * The body type: static, kinematic, or dynamic.
	 * Note: if a dynamic body would have zero mass, the mass is set to one.
	 */
	var type: BodyType = BodyType.STATIC,

	/**
	 * Does this body start out enabled?
	 */
	var enabled: Boolean = true,

	/**
	 * Use this to store application specific body data.
	 */
	var userData: Any? = null,

	/**
	 * Scale the gravity applied to this body.
	 */
	var gravityScale: Double = 1.0,
) {
	fun validate() {
		require(angularVelocity.isFinite()) { "Angular velocity is infinite" }
		require(linearDamping.isFinite()) { "Linear damping is infinite" }
		require(angularDamping.isFinite()) { "Angular damping is infinite" }
		require(angle.isFinite()) { "Angular velocity is infinite" }
		require(gravityScale.isFinite()) { "Gravity scale is infinite" }
		require(position.isFinite) { "Position is infinite" }

		require(!angularVelocity.isNaN()) { "Angular velocity is NaN" }
		require(!linearDamping.isNaN()) { "Linear damping is NaN" }
		require(!angularDamping.isNaN()) { "Angular damping is NaN" }
		require(!angle.isNaN()) { "Angular velocity is NaN" }
		require(!gravityScale.isNaN()) { "Gravity scale is NaN" }

		require(angularDamping >= 0.0) { "Angular damping must be non negative, $angularDamping given" }
		require(linearDamping >= 0.0) { "Linear damping must be non negative, $linearDamping given" }
	}
}

enum class BodyFlags(val bitmask: Int) {
	ISLAND(0x1),
	AWAKE(0x2),
	AUTO_SLEEP(0x4),
	BULLET(0x8),
	FIXED_ROTATION(0x10),
	ENABLED(0x20),
	TOI(0x40);

	fun isit(other: Int): Boolean {
		return (other and bitmask) == bitmask
	}

	fun update(other: Int, state: Boolean): Int {
		if (state)
			return or(other)
		else
			return not(other)
	}

	fun and(other: Int): Int {
		return other and bitmask
	}

	fun or(other: Int): Int {
		return other or bitmask
	}

	fun not(other: Int): Int {
		return other and bitmask.inv()
	}
}
