package ru.dbotthepony.kbox2d.api

import ru.dbotthepony.kvector.util2d.AABB
import ru.dbotthepony.kvector.vector.ndouble.Vector2d

data class MassData(
	val mass: Double,
	val center: Vector2d,
	val inertia: Double,
)


/**
 * A shape is used for collision detection. You can create a shape however you like.
 * Shapes used for simulation in b2World are created automatically when a b2Fixture
 * is created. Shapes may encapsulate a one or more child shapes.
 */
interface IShape<S : IShape<S>> {
	enum class Type {
		CIRCLE,
		EDGE,
		POLYGON,
		CHAIN,
	}

	/**
	 * Clone the concrete shape.
	 */
	fun copy(): S

	/**
	 * Get the type of this shape. You can use this to down cast to the concrete shape.
	 * @return the shape type.
	 */
	val type: Type

	/**
	 * Get the number of child primitives.
	 */
	val childCount: Int


	/**
	 * Test a point for containment in this shape. This only works for convex shapes.
	 * @param xf the shape world transform.
	 * @param p a point in world coordinates.
	 */
	fun testPoint(transform: Transform, p: Vector2d): Boolean { return false }

	/**
	 * Cast a ray against a child shape.
	 * @param output the ray-cast results.
	 * @param input the ray-cast input parameters.
	 * @param transform the transform to be applied to the shape.
	 * @param childIndex the child shape index
	 */
	fun rayCast(input: RayCastInput, transform: Transform, childIndex: Int): RayCastOutput

	/**
	 * Given a transform, compute the associated axis aligned bounding box for a child shape.
	 * @param aabb returns the axis aligned box.
	 * @param xf the world transform of the shape.
	 * @param childIndex the child shape
	 */
	fun computeAABB(transform: Transform, childIndex: Int): AABB

	/**
	 * Compute the mass properties of this shape using its dimensions and density.
	 * The inertia tensor is computed about the local origin.
	 * @param massData returns the mass data for this shape.
	 * @param density the density in kilograms per meter squared.
	 */
	fun computeMass(density: Double): MassData

	/**
	 * Radius of a shape. For polygonal shapes this must be b2_polygonRadius. There is no support for
	 * making rounded polygons.
	 */
	var radius: Double
}
