package ru.dbotthepony.kbox2d.collision.handler

import ru.dbotthepony.kbox2d.api.*
import ru.dbotthepony.kbox2d.api.b2Mul
import ru.dbotthepony.kbox2d.api.b2MulT
import ru.dbotthepony.kbox2d.collision.b2ClipSegmentToLine
import ru.dbotthepony.kbox2d.collision.shapes.CircleShape
import ru.dbotthepony.kbox2d.collision.shapes.EdgeShape
import ru.dbotthepony.kbox2d.collision.shapes.PolygonShape
import ru.dbotthepony.kvector.vector.ndouble.Vector2d
import ru.dbotthepony.kvector.vector.ndouble.times
import java.util.*
import kotlin.collections.ArrayList

internal fun b2CollideEdgeAndCircle(
	edgeA: EdgeShape,
	xfA: Transform,
	circleB: CircleShape,
	xfB: Transform
): Manifold {
	// Compute circle in frame of edge
	val Q = b2MulT(xfA, b2Mul(xfB, circleB.p))

	val A = edgeA.vertex1
	val B = edgeA.vertex2
	val e = B - A

	// Normal points to the right for a CCW winding
	var n = Vector2d(e.y, -e.x)
	val offset = b2Dot(n, Q - A)

	if (edgeA.oneSided && offset < 0.0) {
		return Manifold.EMPTY
	}

	// Barycentric coordinates
	val u = b2Dot(e, B - Q)
	val v = b2Dot(e, Q - A)

	val radius = edgeA.radius + circleB.radius
	val cf = ContactFeature(
		indexB = 0,
		typeB = ContactFeature.Type.VERTEX
	)

	// Region A
	if (v <= 0.0) {
		val P = A
		val d = Q - P
		val dd = b2Dot(d, d)
		if (dd > radius * radius) {
			return Manifold.EMPTY
		}

		// Is there an edge connected to A?
		if (edgeA.oneSided) {
			val A1 = edgeA.vertex0
			val B1 = A
			val e1 = B1 - A1
			val u1 = b2Dot(e1, B1 - Q)

			// Is the circle in Region AB of the previous edge?
			if (u1 > 0.0) {
				return Manifold.EMPTY
			}
		}

		cf.indexA = 0
		cf.typeA = ContactFeature.Type.VERTEX

		return Manifold(
			type = Manifold.Type.CIRCLES,
			localPoint = P,
			points = listOf(
				ManifoldPoint(
					id = ContactID(key = 0, cf = cf),
					localPoint = circleB.p
				)
			)
		)
	}

	// Region B
	if (u <= 0.0) {
		val P = B
		val d = Q - P
		val dd = b2Dot(d, d)
		if (dd > radius * radius) {
			return Manifold.EMPTY
		}

		// Is there an edge connected to B?
		if (edgeA.oneSided) {
			val B2 = edgeA.vertex3
			val A2 = B
			val e2 = B2 - A2
			val v2 = b2Dot(e2, Q - A2)

			// Is the circle in Region AB of the next edge?
			if (v2 > 0.0) {
				return Manifold.EMPTY
			}
		}

		cf.indexA = 1
		cf.typeA = ContactFeature.Type.VERTEX

		return Manifold(
			type = Manifold.Type.CIRCLES,
			localPoint = P,
			points = listOf(
				ManifoldPoint(
					id = ContactID(key = 0, cf = cf),
					localPoint = circleB.p
				)
			)
		)
	}

	// Region AB
	val den = b2Dot(e, e)
	check(den > 0.0) { den }
	val P = (1.0 / den) * (u * A + v * B)
	val d = Q - P
	val dd = b2Dot(d, d)
	if (dd > radius * radius) {
		return Manifold.EMPTY
	}

	if (offset < 0.0) {
		n = Vector2d(-n.x, -n.y)
	}

	n = n.normalized

	cf.indexA = 0
	cf.typeA = ContactFeature.Type.FACE

	return Manifold(
		type = Manifold.Type.FACE_A,
		localNormal = n,
		localPoint = A,
		points = listOf(
			ManifoldPoint(
				id = ContactID(key = 0, cf = cf),
				localPoint = circleB.p
			)
		)
	)
}

// This structure is used to keep track of the best separating axis.
private data class EPAxis(
	var normal: Vector2d = Vector2d.ZERO,
	var type: Type = Type.UNKNOWN,
	var index: Int = -1,
	var separation: Double = -Double.MAX_VALUE
) {
	enum class Type {
		UNKNOWN,
		EDGE_A,
		EDGE_B
	}
}

// This holds polygon B expressed in frame A.
private data class TempPolygon(
	val vertices: ArrayList<Vector2d> = ArrayList(),
	val normals: ArrayList<Vector2d> = ArrayList(),
)

// Reference face used for clipping
private data class ReferenceFace(
	var i1: Int = 0,
	var i2: Int = 0,
	var v1: Vector2d = Vector2d.ZERO,
	var v2: Vector2d = Vector2d.ZERO,
	var normal: Vector2d = Vector2d.ZERO,

	var sideNormal1: Vector2d = Vector2d.ZERO,
	var sideOffset1: Double = 0.0,

	var sideNormal2: Vector2d = Vector2d.ZERO,
	var sideOffset2: Double = 0.0,
)

// Reference face used for clipping
private fun b2ComputeEdgeSeparation(
	polygonB: TempPolygon,
	v1: Vector2d,
	normal1: Vector2d
): EPAxis {
	val axis = EPAxis(type = EPAxis.Type.EDGE_A)
	val axes = arrayOf(normal1, -normal1)

	// Find axis with least overlap (min-max problem)
	for (j in axes.indices) {
		var sj = Double.MAX_VALUE

		// Find deepest polygon vertex along axis j
		for (v in polygonB.vertices) {
			val si = b2Dot(axes[j], v - v1)

			if (si < sj) {
				sj = si
			}
		}

		if (sj > axis.separation) {
			axis.index = j
			axis.separation = sj
			axis.normal = axes[j]
		}
	}

	return axis
}

private fun b2ComputePolygonSeparation(
	polygonB: TempPolygon,
	v1: Vector2d,
	v2: Vector2d
): EPAxis {
	val axis = EPAxis()

	for (i in polygonB.vertices.indices) {
		val n = -polygonB.normals[i]

		val s1 = b2Dot(n, polygonB.vertices[i] - v1)
		val s2 = b2Dot(n, polygonB.vertices[i] - v2)
		val s = b2Min(s1, s2)

		if (s > axis.separation) {
			axis.type = EPAxis.Type.EDGE_B
			axis.index = i
			axis.separation = s
			axis.normal = n
		}
	}

	return axis
}

private const val k_relativeTol = 0.98
private const val k_absoluteTol = 0.001
private const val sinTol = 0.1

internal fun b2CollideEdgeAndPolygon(
	edgeA: EdgeShape,
	xfA: Transform,
	polygonB: PolygonShape,
	xfB: Transform
): Manifold {
	val xf = b2MulT(xfA, xfB)

	val centroidB = b2Mul(xf, polygonB.centroid)

	val v1 = edgeA.vertex1
	val v2 = edgeA.vertex2

	val edge1 = (v2 - v1).normalized

	// Normal points to the right for a CCW winding
	var normal1 = Vector2d(edge1.y, -edge1.x)
	val offset1 = b2Dot(normal1, centroidB - v1)

	if (edgeA.oneSided && offset1 < 0.0) {
		return Manifold.EMPTY
	}

	// Get polygonB in frameA
	val tempPolygonB = TempPolygon(ArrayList(polygonB.count), ArrayList(polygonB.count))

	for (i in 0 until polygonB.count) {
		tempPolygonB.vertices.add(b2Mul(xf, polygonB.vertices[i]))
		tempPolygonB.normals.add(b2Mul(xf.q, polygonB.normals[i]))
	}

	val radius = polygonB.radius + edgeA.radius

	val edgeAxis = b2ComputeEdgeSeparation(tempPolygonB, v1, normal1)
	if (edgeAxis.separation > radius) {
		return Manifold.EMPTY
	}

	val polygonAxis = b2ComputePolygonSeparation(tempPolygonB, v1, v2)
	if (polygonAxis.separation > radius) {
		return Manifold.EMPTY
	}

	// Use hysteresis for jitter reduction.
	var primaryAxis: EPAxis

	if (polygonAxis.separation - radius > k_relativeTol * (edgeAxis.separation - radius) + k_absoluteTol) {
		primaryAxis = polygonAxis
	} else {
		primaryAxis = edgeAxis
	}

	if (edgeA.oneSided) {
		// Smooth collision
		// See https://box2d.org/posts/2020/06/ghost-collisions/

		val edge0 = (v1 - edgeA.vertex0).normalized
		val normal0 = Vector2d(edge0.y, -edge0.x)
		val convex1 = b2Cross(edge0, edge1) >= 0.0f

		val edge2 = (edgeA.vertex3 - v2).normalized
		val normal2 = Vector2d(edge2.y, -edge2.x)
		val convex2 = b2Cross(edge1, edge2) >= 0.0f

		val side1 = b2Dot(primaryAxis.normal, edge1) <= 0.0

		// Check Gauss Map
		if (side1) {
			if (convex1) {
				if (b2Cross(primaryAxis.normal, normal0) > sinTol) {
					// Skip region
					return Manifold.EMPTY
				}

				// Admit region
			} else {
				// Snap region
				primaryAxis = edgeAxis
			}
		} else {
			if (convex2) {
				if (b2Cross(normal2, primaryAxis.normal) > sinTol) {
					// Skip region
					return Manifold.EMPTY
				}

				// Admit region
			} else {
				// Snap region
				primaryAxis = edgeAxis
			}
		}
	}

	val ref = ReferenceFace()
	var manifoldType: Manifold.Type
	val clipPoints = Array(2) { ClipVertex() }

	if (primaryAxis.type == EPAxis.Type.EDGE_A) {
		manifoldType = Manifold.Type.FACE_A

		// Search for the polygon normal that is most anti-parallel to the edge normal.
		var bestIndex = 0
		var bestValue = b2Dot(primaryAxis.normal, tempPolygonB.normals[0])

		for (i in 1 until tempPolygonB.normals.size) {
			val value = b2Dot(primaryAxis.normal, tempPolygonB.normals[i])

			if (value < bestValue) {
				bestValue = value
				bestIndex = i
			}
		}

		val i1 = bestIndex
		val i2 = if (i1 + 1 < tempPolygonB.normals.size) i1 + 1 else 0

		clipPoints[0].v = tempPolygonB.vertices[i1]
		clipPoints[0].id.cf.indexA = 0
		clipPoints[0].id.cf.indexB = i1
		clipPoints[0].id.cf.typeA = ContactFeature.Type.FACE
		clipPoints[0].id.cf.typeB = ContactFeature.Type.VERTEX

		clipPoints[1].v = tempPolygonB.vertices[i2]
		clipPoints[1].id.cf.indexA = 0
		clipPoints[1].id.cf.indexB = i2
		clipPoints[1].id.cf.typeA = ContactFeature.Type.FACE
		clipPoints[1].id.cf.typeB = ContactFeature.Type.VERTEX

		ref.i1 = 0
		ref.i2 = 1
		ref.v1 = v1
		ref.v2 = v2
		ref.normal = primaryAxis.normal
		ref.sideNormal1 = -edge1
		ref.sideNormal2 = edge1
	} else {
		manifoldType = Manifold.Type.FACE_B

		clipPoints[0].v = v2
		clipPoints[0].id.cf.indexA = 1
		clipPoints[0].id.cf.indexB = primaryAxis.index
		clipPoints[0].id.cf.typeA = ContactFeature.Type.VERTEX
		clipPoints[0].id.cf.typeB = ContactFeature.Type.FACE

		clipPoints[1].v = v1
		clipPoints[1].id.cf.indexA = 0
		clipPoints[1].id.cf.indexB = primaryAxis.index
		clipPoints[1].id.cf.typeA = ContactFeature.Type.VERTEX
		clipPoints[1].id.cf.typeB = ContactFeature.Type.FACE

		ref.i1 = primaryAxis.index;
		ref.i2 = if (ref.i1 + 1 < tempPolygonB.normals.size) ref.i1 + 1 else 0
		ref.v1 = tempPolygonB.vertices[ref.i1]
		ref.v2 = tempPolygonB.vertices[ref.i2]
		ref.normal = tempPolygonB.normals[ref.i1]

		// CCW winding
		ref.sideNormal1 = Vector2d(ref.normal.y, -ref.normal.x)
		ref.sideNormal2 = -ref.sideNormal1
	}

	ref.sideOffset1 = b2Dot(ref.sideNormal1, ref.v1)
	ref.sideOffset2 = b2Dot(ref.sideNormal2, ref.v2)

	// Clip incident edge against reference face side planes

	// Clip to side 1
	val clipPoints1 = b2ClipSegmentToLine(clipPoints, ref.sideNormal1, ref.sideOffset1, ref.i1)

	if (clipPoints1.size < b2_maxManifoldPoints) {
		return Manifold.EMPTY
	}

	// Clip to side 2
	val clipPoints2 = b2ClipSegmentToLine(clipPoints1, ref.sideNormal2, ref.sideOffset2, ref.i2)

	if (clipPoints2.size < b2_maxManifoldPoints) {
		return Manifold.EMPTY
	}

	val localNormal: Vector2d
	val localPoint: Vector2d

	// Now clipPoints2 contains the clipped points.
	if (primaryAxis.type == EPAxis.Type.EDGE_A) {
		localNormal = ref.normal
		localPoint = ref.v1
	} else {
		localNormal = polygonB.normals[ref.i1]
		localPoint = polygonB.vertices[ref.i1]
	}

	val points = ArrayList<ManifoldPoint>(b2_maxManifoldPoints)

	for (i in 0 until b2_maxManifoldPoints) {
		val separation = b2Dot(ref.normal, clipPoints2[i].v - ref.v1)

		if (separation <= radius) {
			val cp = ManifoldPoint()

			if (primaryAxis.type == EPAxis.Type.EDGE_A) {
				cp.localPoint = b2MulT(xf, clipPoints2[i].v)
				cp.id = clipPoints2[i].id
			} else {
				cp.localPoint = clipPoints2[i].v
				cp.id.cf.typeA = clipPoints2[i].id.cf.typeB
				cp.id.cf.typeB = clipPoints2[i].id.cf.typeA
				cp.id.cf.indexA = clipPoints2[i].id.cf.indexB
				cp.id.cf.indexB = clipPoints2[i].id.cf.indexA
			}

			points.add(cp)
		}
	}

	return Manifold(
		localNormal = localNormal,
		localPoint = localPoint,
		type = manifoldType,
		points = Collections.unmodifiableList(points))
}
