package ru.dbotthepony.kbox2d.dynamics.internal

import ru.dbotthepony.kbox2d.api.*
import ru.dbotthepony.kbox2d.collision.DistanceProxy
import ru.dbotthepony.kbox2d.collision.b2Distance
import ru.dbotthepony.kbox2d.dynamics.contact.AbstractContact
import ru.dbotthepony.kbox2d.dynamics.joint.AbstractJoint
import ru.dbotthepony.kbox2d.dynamics.B2Body
import ru.dbotthepony.kvector.vector.ndouble.Vector2d
import ru.dbotthepony.kvector.vector.ndouble.times
import java.util.*
import kotlin.collections.ArrayList
import kotlin.math.absoluteValue

private const val linTolSqr = b2_linearSleepTolerance * b2_linearSleepTolerance
private const val angTolSqr = b2_angularSleepTolerance * b2_angularSleepTolerance

private const val checkPositions = false

/*
Position Correction Notes
=========================
I tried the several algorithms for position correction of the 2D revolute joint.
I looked at these systems:
- simple pendulum (1m diameter sphere on massless 5m stick) with initial angular velocity of 100 rad/s.
- suspension bridge with 30 1m long planks of length 1m.
- multi-link chain with 30 1m long links.

Here are the algorithms:

Baumgarte - A fraction of the position error is added to the velocity error. There is no
separate position solver.

Pseudo Velocities - After the velocity solver and position integration,
the position error, Jacobian, and effective mass are recomputed. Then
the velocity constraints are solved with pseudo velocities and a fraction
of the position error is added to the pseudo velocity error. The pseudo
velocities are initialized to zero and there is no warm-starting. After
the position solver, the pseudo velocities are added to the positions.
This is also called the First Order World method or the Position LCP method.

Modified Nonlinear Gauss-Seidel (NGS) - Like Pseudo Velocities except the
position error is re-computed for each constraint and the positions are updated
after the constraint is solved. The radius vectors (aka Jacobians) are
re-computed too (otherwise the algorithm has horrible instability). The pseudo
velocity states are not needed because they are effectively zero at the beginning
of each iteration. Since we have the current position error, we allow the
iterations to terminate early if the error becomes smaller than b2_linearSlop.

Full NGS or just NGS - Like Modified NGS except the effective mass are re-computed
each time a constraint is solved.

Here are the results:
Baumgarte - this is the cheapest algorithm but it has some stability problems,
especially with the bridge. The chain links separate easily close to the root
and they jitter as they struggle to pull together. This is one of the most common
methods in the field. The big drawback is that the position correction artificially
affects the momentum, thus leading to instabilities and false bounce. I used a
bias factor of 0.2. A larger bias factor makes the bridge less stable, a smaller
factor makes joints and contacts more spongy.

Pseudo Velocities - the is more stable than the Baumgarte method. The bridge is
stable. However, joints still separate with large angular velocities. Drag the
simple pendulum in a circle quickly and the joint will separate. The chain separates
easily and does not recover. I used a bias factor of 0.2. A larger value lead to
the bridge collapsing when a heavy cube drops on it.

Modified NGS - this algorithm is better in some ways than Baumgarte and Pseudo
Velocities, but in other ways it is worse. The bridge and chain are much more
stable, but the simple pendulum goes unstable at high angular velocities.

Full NGS - stable in all tests. The joints display good stiffness. The bridge
still sags, but this is better than infinite forces.

Recommendations
Pseudo Velocities are not really worthwhile because the bridge and chain cannot
recover from joint separation. In other cases the benefit over Baumgarte is small.

Modified NGS is not a robust method for the revolute joint due to the violent
instability seen in the simple pendulum. Perhaps it is viable with other constraint
types, especially scalar constraints where the effective mass is a scalar.

This leaves Baumgarte and Full NGS. Baumgarte has small, but manageable instabilities
and is very fast. I don't think we can escape Baumgarte, especially in highly
demanding cases where high constraint fidelity is not needed.

Full NGS is robust and easy on the eyes. I recommend this as an option for
higher fidelity simulation and certainly for suspension bridges and long chains.
Full NGS might be a good choice for ragdolls, especially motorized ragdolls where
joint separation can be problematic. The number of NGS iterations can be reduced
for better performance without harming robustness much.

Each joint in a can be handled differently in the position solver. So I recommend
a system where the user can select the algorithm on a per joint basis. I would
probably default to the slower Full NGS and let the user select the faster
Baumgarte method in performance critical scenarios.
*/

/*
Cache Performance

The Box2D solvers are dominated by cache misses. Data structures are designed
to increase the number of cache hits. Much of misses are due to random access
to body data. The constraint structures are iterated over linearly, which leads
to few cache misses.

The bodies are not accessed during iteration. Instead read only data, such as
the mass values are stored with the constraints. The mutable data are the constraint
impulses and the bodies velocities/positions. The impulses are held inside the
constraint structures. The body velocities/positions are held in compact, temporary
arrays to increase the number of cache hits. Linear and angular velocity are
stored in a single array since multiple arrays lead to multiple misses.
*/

/*
2D Rotation

R = [cos(theta) -sin(theta)]
    [sin(theta) cos(theta) ]

thetaDot = omega

Let q1 = cos(theta), q2 = sin(theta).
R = [q1 -q2]
    [q2  q1]

q1Dot = -thetaDot * q2
q2Dot = thetaDot * q1

q1_new = q1_old - dt * w * q2
q2_new = q2_old + dt * w * q1
then normalize.

This might be faster than computing sin+cos.
However, we can compute sin+cos of the same angle fast.
*/

/**
 * This is an internal class.
 */
internal class Island(
	initialBodyCapacity: Int = 0,
	initialContactCapacity: Int = 0,
	initialJointCapacity: Int = 0,
	val listener: IContactListener? = null
) {
	private val bodies = ArrayList<B2Body>(initialBodyCapacity)
	private val contacts = ArrayList<AbstractContact>(initialContactCapacity)
	private val joints = ArrayList<AbstractJoint>(initialJointCapacity)

	private val velocities = ArrayList<ru.dbotthepony.kbox2d.api.B2Velocity>(initialBodyCapacity)
	private val positions = ArrayList<ru.dbotthepony.kbox2d.api.B2Position>(initialBodyCapacity)

	val bodiesAccess: List<B2Body> = Collections.unmodifiableList(bodies)

	fun clear() {
		bodies.clear()
		contacts.clear()
		joints.clear()

		velocities.clear()
		positions.clear()
	}

	fun add(body: B2Body) {
		body.islandIndex = bodies.size
		bodies.add(body)
		velocities.add(ru.dbotthepony.kbox2d.api.B2Velocity())
		positions.add(ru.dbotthepony.kbox2d.api.B2Position())
	}

	fun add(contact: AbstractContact) {
		contacts.add(contact)
	}

	fun add(joint: AbstractJoint) {
		joints.add(joint)
	}

	fun solve(profile: ru.dbotthepony.kbox2d.api.ProfileData, step: ru.dbotthepony.kbox2d.api.B2TimeStep, gravity: Vector2d, allowSleep: Boolean) {
		val h = step.dt

		// Integrate velocities and apply damping. Initialize the body state.
		for ((i, body) in bodies.withIndex()) {
			val c = body.sweep.c
			val a = body.sweep.a
			var v = body.linearVelocity
			var w = body.angularVelocity

			// Store positions for continuous collision.
			body.sweep.c0 = body.sweep.c
			body.sweep.a0 = body.sweep.a

			if (body.type == ru.dbotthepony.kbox2d.api.BodyType.DYNAMIC) {
				// Integrate velocities.
				v += (gravity * body.gravityScale * body.mass + body.force) * body.invMass * h
				w += h * body.rotInertiaInv * body.torque

				// Apply damping.
				// ODE: dv/dt + c * v = 0
				// Solution: v(t) = v0 * exp(-c * t)
				// Time step: v(t + dt) = v0 * exp(-c * (t + dt)) = v0 * exp(-c * t) * exp(-c * dt) = v * exp(-c * dt)
				// v2 = exp(-c * dt) * v1
				// Pade approximation:
				// v2 = v1 * 1 / (1 + c * dt)
				v *= 1.0 / (1.0 + h * body.linearDamping)
				w *= 1.0 / (1.0 + h * body.angularDamping)
			}

			positions[i].c = c
			positions[i].a = a
			velocities[i].v = v
			velocities[i].w = w
		}

		var timer = System.nanoTime()

		// Solver data
		val solverData = ru.dbotthepony.kbox2d.api.B2SolverData(
			step = step,
			positions = positions,
			velocities = velocities
		)

		// Initialize velocity constraints.
		val contactSolver = ContactSolver(
			step = step,
			contacts = contacts,
			positions = positions,
			velocities = velocities,
		)

		contactSolver.initializeVelocityConstraints()

		if (step.warmStarting) {
			contactSolver.warmStart()
		}

		for (joint in joints) {
			joint.initVelocityConstraints(solverData)
		}

		profile.solveInit = System.nanoTime() - timer
		timer = System.nanoTime()

		// Solve velocity constraints
		for (i in 0 until step.velocityIterations) {
			for (joint in joints) {
				joint.solveVelocityConstraints(solverData)
			}

			contactSolver.solveVelocityConstraints()
		}

		// Store impulses for warm starting
		contactSolver.storeImpulses()
		profile.solveVelocity = System.nanoTime() - timer
		timer = System.nanoTime()

		// Integrate positions
		for (i in bodies.indices) {
			var c = positions[i].c
			var a = positions[i].a
			var v = velocities[i].v
			var w = velocities[i].w

			// Check for large velocities
			val translation = h * v
			if (translation.dot(translation) > b2_maxTranslationSquared) {
				v *= b2_maxTranslation / translation.length
			}

			val rotation = h * w
			if (rotation * rotation > b2_maxRotationSquared) {
				w *= b2_maxRotation / rotation.absoluteValue
			}

			// Integrate
			c += h * v
			a += h * w

			positions[i].c = c
			positions[i].a = a
			velocities[i].v = v
			velocities[i].w = w
		}

		profile.integratePositions = System.nanoTime() - timer
		timer = System.nanoTime()

		// Solve position constraints
		var positionSolved = false

		for (i in 0 until step.positionIterations) {
			val contactsOkay = contactSolver.solvePositionConstraints()
			var jointsOkay = true

			for (joint in joints) {
				val jointOkay = joint.solvePositionConstraints(solverData)
				jointsOkay = jointsOkay && jointOkay
			}

			if (contactsOkay && jointsOkay) {
				// Exit early if the position errors are small.
				positionSolved = true
				break
			}
		}

		// Copy state buffers back to the bodies
		for ((i, body) in bodies.withIndex()) {
			body.sweep.c = positions[i].c
			body.sweep.a = positions[i].a
			body.linearVelocity = velocities[i].v
			body.angularVelocity = velocities[i].w
			body.synchronizeTransform()
		}

		profile.solvePosition = System.nanoTime() - timer

		report(contactSolver.velocityConstraints)

		if (allowSleep) {
			var minSleepTime = Double.MAX_VALUE

			for (body in bodies) {
				if (body.type == ru.dbotthepony.kbox2d.api.BodyType.STATIC)
					continue

				if (
					!body.allowAutoSleep ||
					body.angularVelocity * body.angularVelocity > angTolSqr ||
					body.linearVelocity.dot(body.linearVelocity) > linTolSqr
				) {
					body.sleepTime = 0.0
					minSleepTime = 0.0
				} else {
					body.sleepTime += h
					minSleepTime = minSleepTime.coerceAtMost(body.sleepTime)
				}
			}

			if (minSleepTime >= b2_timeToSleep && positionSolved) {
				for (body in bodies) {
					body.isAwake = false
				}
			}
		}
	}

	fun solveTOI(subStep: ru.dbotthepony.kbox2d.api.B2TimeStep, toiIndexA: Int, toiIndexB: Int) {
		check(toiIndexA < bodies.size) { "$toiIndexA >= ${bodies.size}" }
		check(toiIndexB < bodies.size) { "$toiIndexB >= ${bodies.size}" }

		// Initialize the body state.
		for ((i, body) in bodies.withIndex()) {
			positions[i].c = body.sweep.c
			positions[i].a = body.sweep.a
			velocities[i].v = body.linearVelocity
			velocities[i].w = body.angularVelocity
		}

		val contactSolver = ContactSolver(
			contacts = contacts,
			step = subStep,
			positions = positions,
			velocities = velocities,
		)

		// Solve position constraints.
		for (i in 0 until subStep.positionIterations) {
			val contactsOkay = contactSolver.solveTOIPositionConstraints(toiIndexA, toiIndexB)

			if (contactsOkay) {
				break
			}
		}

		// Is the new position really safe?
		if (checkPositions) {
			for (c in contacts) {
				val fA = c.fixtureA
				val fB = c.fixtureB

				val bA = fA.body!!
				val bB = fB.body!!

				val indexA = c.childIndexA
				val indexB = c.childIndexB

				val cache = SimplexCache()

				val output = b2Distance(
					cache = cache,
					proxyA = DistanceProxy(fA.shape, indexA),
					proxyB = DistanceProxy(fB.shape, indexB),
					transformA = bA.transform,
					transformB = bB.transform,
				)

				if (output.distance == 0.0 || output.newCache.count == 3) {
					// cache.count += 0;
					// doesn't make much sense
				}
			}
		}

		// Leap of faith to new safe state.
		bodies[toiIndexA].sweep.c0 = positions[toiIndexA].c
		bodies[toiIndexA].sweep.a0 = positions[toiIndexA].a
		bodies[toiIndexB].sweep.c0 = positions[toiIndexB].c
		bodies[toiIndexB].sweep.a0 = positions[toiIndexB].a

		// No warm starting is needed for TOI events because warm
		// starting impulses were applied in the discrete solver.
		contactSolver.initializeVelocityConstraints()

		// Solve velocity constraints.
		for (i in 0 until subStep.velocityIterations) {
			contactSolver.solveVelocityConstraints()
		}

		// Don't store the TOI contact forces for warm starting
		// because they can be quite large.

		val h = subStep.dt

		// Integrate positions
		for ((i, body) in bodies.withIndex()) {
			var c = positions[i].c
			var a = positions[i].a
			var v = velocities[i].v
			var w = velocities[i].w

			// Check for large velocities
			val translation = h * v
			if (translation.dot(translation) > b2_maxTranslationSquared) {
				v *= b2_maxTranslation / translation.length
			}

			val rotation = h * w
			if (rotation * rotation > b2_maxRotationSquared) {
				w *= b2_maxRotation / rotation.absoluteValue
			}

			// Integrate
			c += h * v
			a += h * w

			positions[i].c = c
			positions[i].a = a
			velocities[i].v = v
			velocities[i].w = w

			// Sync bodies
			body.sweep.c = c
			body.sweep.a = a
			body.linearVelocity = v
			body.angularVelocity = w
			body.synchronizeTransform()
		}

		report(contactSolver.velocityConstraints)
	}

	fun report(constraints: List<ContactVelocityConstraint>) {
		val listener = listener ?: return

		for ((i, contact) in contacts.withIndex()) {
			val vc = constraints[i]

			val impulse = ContactImpulse(
				normalImpulses = DoubleArray(vc.points.size) { vc.points[it].normalImpulse },
				tangentImpulses = DoubleArray(vc.points.size) { vc.points[it].tangentImpulse },
			)

			listener.postSolve(contact, impulse)
		}
	}
}
