
@file:Suppress("unused")

package ru.dbotthepony.kvector.matrix.ndouble

import ru.dbotthepony.kvector.api.*
import ru.dbotthepony.kvector.api.concrete.IMatrix2d
import ru.dbotthepony.kvector.narray.Double2Dimensional
import ru.dbotthepony.kvector.vector.ndouble.MutableVector2d
import ru.dbotthepony.kvector.vector.ndouble.Vector2d

/**
 * Represents immutable concrete matrix with 2x2 dimensions, storing values as [Double]s.
 *
 * See [Matrix4d] for documentation, with reflection of this class.
 *
 * Vectorized access use [Vector2d], generic interface is represented by [IMatrix2d].
 *
 * It does *not* implement [ISquareMatrix] since it doesn't make much sense.
 */
class Matrix2d : AbstractMatrixVd<Matrix2d>, IMatrix2d<Matrix2d> {
	constructor(
		c00: Double = 1.0, c01: Double = 0.0,
		c10: Double = 0.0, c11: Double = 1.0,
	) : super(2, 2, Double2Dimensional(2, 2).also {
		it[0, 0] = c00
		it[1, 0] = c10

		it[0, 1] = c01
		it[1, 1] = c11
	})

	private constructor(memory: Double2Dimensional) : super(2, 2, memory)

	override val flexible: Boolean = false

	override fun factorize(memory: Double2Dimensional): Matrix2d {
		requireSizeEquals(memory)
		return Matrix2d(memory)
	}

	override val c00: Double get() = memory[0, 0]
	override val c10: Double get() = memory[1, 0]
	override val c01: Double get() = memory[0, 1]
	override val c11: Double get() = memory[1, 1]

	override val r00 by this::c00
	override val r01 by this::c10
	override val r10 by this::c01
	override val r11 by this::c11

	override val a11 by this::c00
	override val a21 by this::c01
	override val a12 by this::c10
	override val a22 by this::c11

	override val b11 by this::c00
	override val b21 by this::c10
	override val b12 by this::c01
	override val b22 by this::c11

	override val c0: Vector2d get() = Vector2d(c00, c01)
	override val c1: Vector2d get() = Vector2d(c10, c11)

	override val r0: Vector2d get() = Vector2d(r00, r01)
	override val r1: Vector2d get() = Vector2d(r10, r11)

	companion object {
		/**
		 * Initializes new matrix, with arguments taken in *Row-Major* order, that is, fist row of values
		 * is fist row in created matrix.
		 */
		fun rm(
			r00: Double = 1.0, r01: Double = 0.0,
			r10: Double = 0.0, r11: Double = 1.0,
		): Matrix2d {
			return Matrix2d(
				c00 = r00, c10 = r01,
				c01 = r10, c11 = r11,
			)
		}

		/**
		 * Initializes new zero matrix.
		 *
		 * Since [Matrix2d] is immutable, it is encouraged you use [ZERO] property instead.
		 */
		fun zero(): Matrix2d {
			return Matrix2d(
				c00 = 0.0, c10 = 0.0,
				c01 = 0.0, c11 = 0.0,
			)
		}

		/**
		 * All values of this matrix are zero
		 */
		@JvmField val ZERO = zero()

		/**
		 * Main diagonal of this matrix is 1s
		 */
		@JvmField val IDENTITY = Matrix2d()
	}
}

/**
 * Represents mutable concrete matrix with 2x2 dimensions, storing values as [Double]s.
 *
 * See [MutableMatrix4d] for documentation, with reflection of this class.
 *
 * Vectorized access use [MutableVector2d], generic interface is represented by [IMatrix2d].
 *
 * It does *not* implement [ISquareMatrix] since it doesn't make much sense.
 */
class MutableMatrix2d : AbstractMutableMatrixVd<MutableMatrix2d>, IMatrix2d<MutableMatrix2d> {
	constructor(
		c00: Double = 1.0, c01: Double = 0.0,
		c10: Double = 0.0, c11: Double = 1.0,
	) : super(2, 2, Double2Dimensional(2, 2).also {
		it[0, 0] = c00
		it[1, 0] = c10

		it[0, 1] = c01
		it[1, 1] = c11
	})

	private constructor(memory: Double2Dimensional) : super(2, 2, memory)

	override val flexible: Boolean = false

	override fun factorize(memory: Double2Dimensional): MutableMatrix2d {
		requireSizeEquals(memory)
		return MutableMatrix2d(memory)
	}

	override var c00: Double
		get() = memory[0, 0]
		set(value) { memory[0, 0] = value }
	override var c10: Double
		get() = memory[1, 0]
		set(value) { memory[1, 0] = value }
	override var c01: Double
		get() = memory[0, 1]
		set(value) { memory[0, 1] = value }
	override var c11: Double
		get() = memory[1, 1]
		set(value) { memory[1, 1] = value }

	override var r00 by this::c00
	override var r01 by this::c10
	override var r10 by this::c01
	override var r11 by this::c11

	override var a11 by this::c00
	override var a21 by this::c01
	override var a12 by this::c10
	override var a22 by this::c11

	override var b11 by this::c00
	override var b21 by this::c10
	override var b12 by this::c01
	override var b22 by this::c11

	override var c0: Vector2d
		get() = object : MutableVector2d(c00, c01) {
			override var x by this@MutableMatrix2d::c00
			override var y by this@MutableMatrix2d::c01
		}
		set(value) {
			c00 = value.x
			c01 = value.y
		}
	override var c1: Vector2d
		get() = object : MutableVector2d(c10, c11) {
			override var x by this@MutableMatrix2d::c10
			override var y by this@MutableMatrix2d::c11
		}
		set(value) {
			c10 = value.x
			c11 = value.y
		}

	override var r0: Vector2d
		get() = object : MutableVector2d(r00, r01) {
			override var x by this@MutableMatrix2d::r00
			override var y by this@MutableMatrix2d::r01
		}
		set(value) {
			r00 = value.x
			r01 = value.y
		}
	override var r1: Vector2d
		get() = object : MutableVector2d(r10, r11) {
			override var x by this@MutableMatrix2d::r10
			override var y by this@MutableMatrix2d::r11
		}
		set(value) {
			r10 = value.x
			r11 = value.y
		}

	companion object {
		/**
		 * Initializes new matrix, with arguments taken in *Row-Major* order, that is, fist row of values
		 * is fist row in created matrix.
		 */
		fun rm(
			r00: Double = 1.0, r01: Double = 0.0,
			r10: Double = 0.0, r11: Double = 1.0,
		): MutableMatrix2d {
			return MutableMatrix2d(
				c00 = r00, c10 = r01,
				c01 = r10, c11 = r11,
			)
		}

		/**
		 * Initializes new zero matrix.
		 */
		fun zero(): MutableMatrix2d {
			return MutableMatrix2d(
				c00 = 0.0, c10 = 0.0,
				c01 = 0.0, c11 = 0.0,
			)
		}
	}
}